import panel as pn

pn.extension(template="fast")

pn.panel("This is a basic Panel app").servable()

pn.state.template.param.update(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Basic App",
)



"""The JSActions component enables the app to trigger methods on the js side"""
import panel as pn
import param


class JSActions(pn.reactive.ReactiveHTML):  # pylint: disable=too-many-ancestors
    """Provides methods to trigger methods on the js side"""

    # .event cannot trigger on js side. Thus I use Integer
    _url = param.String()
    _open = param.Boolean()
    _copy = param.Boolean()

    _template = """<div id="jsaction" style="height:0px;width:0px"></div>"""
    _scripts = {
        # pylint: disable=line-too-long
        "_open": """
if (data._url.startsWith("http")){
    url=data._url
} else {
    url=window.location.href.substring(0, location.href.lastIndexOf('/')+1) +  data._url;
}
window.open(url, '_blank')""",
        "_copy": """
if (data._url.startsWith("http")){
    url=data._url
} else {
    url=window.location.href.substring(0, location.href.lastIndexOf('/')+1) +  data._url;
}
navigator.clipboard.writeText(url)
""",
        # pylint: enable=line-too-long
    }

    def __init__(self):
        super().__init__(height=0, width=0, sizing_mode="fixed", margin=0)

    def open(self, url: str):
        """Opens the url in a new tab"""
        if url:
            self._url = url
            self._open = not self._open
        else:
            raise ValueError("No url to open")

    def copy(self, url: str):
        """Copies the url to the clipboard"""
        if url:
            self._url = url
            self._copy = not self._copy
        else:
            raise ValueError("No url to copy")



"""Provides an iframe component to embed another web page in your page"""

import panel as pn
import param

from panel_sharing import config

loading_indicator = pn.indicators.LoadingSpinner(
    value=True, width=100, height=100, sizing_mode="fixed"
)


def iframe_string(src: str) -> str:
    """Returns an iframe string"""
    # pylint: disable=line-too-long
    return f"""<iframe frameborder="0" title="panel app" style="width: 100%;height:100%";flex-grow: 1" src="{src}" allow="accelerometer;autoplay;camera;document-domain;encrypted-media;fullscreen;gamepad;geolocation;gyroscope;layout-animations;legacy-image-formats;microphone;oversized-images;payment;publickey-credentials-get;speaker-selection;sync-xhr;unoptimized-images;unsized-media;screen-wake-lock;web-share;xr-spatial-tracking"></iframe>"""


def _iframe(src: str, sizing_mode="stretch_both", **params) -> pn.pane.HTML:
    """Returns an iframe inside a HTML pane

    Use this to embed another web page inside your app.

    Args:
        src: The url of the web page to embed
        sizing_mode: Defaults to "stretch_both".

    Returns:
        An iframe inside a HTML pane
    """
    if not src:
        return loading_indicator
    # pylint: disable=line-too-long
    return pn.pane.HTML(iframe_string(src=src), sizing_mode=sizing_mode, **params)


def iframe(src: param.String):
    """An iframe that embeds the `src` url."""
    pane = pn.bind(_iframe, src=src)
    return pn.Column(pn.panel(pane, sizing_mode="stretch_both"), sizing_mode="stretch_both")


faq = pn.pane.Markdown(config.FAQ, name="FAQ", sizing_mode="stretch_both")
about = pn.pane.Markdown(config.ABOUT, name="ABOUT", sizing_mode="stretch_both")



"""The BuildProject component enables users to convert the project"""
import panel as pn
import param

from panel_sharing.components.js_actions import JSActions
from panel_sharing.utils import del_query_params


class ProjectBuilder(pn.viewable.Viewer):
    """A component that enables users to convert the project"""

    convert = param.Event()

    open_developer_link = param.Event()

    _state = param.Parameter()
    jsactions = param.ClassSelector(class_=JSActions)

    def __init__(self, state, **params):
        if "jsactions" not in params:
            params["jsactions"] = JSActions()
        super().__init__(_state=state, **params)

        self._panel = self._get_panel()

    def __panel__(self):
        return self._panel

    @pn.depends("convert", watch=True)
    def _convert(self):
        self._state.build()

        if pn.state.notifications:
            pn.state.notifications.success("Build succeeded")

        del_query_params()
        pn.state.location.update_query(project=self._state.project.to_base64())

    @pn.depends("open_developer_link", watch=True)
    def _open_developer_link(self):
        self.jsactions.open(url=self._state.development_url)

    def _download_callback(self):
        return self._state.project.to_zipped_folder()

    def _get_panel(self):
        self.convert_button = pn.widgets.Button.from_param(
            self.param.convert,
            name="Convert",
            sizing_mode="stretch_width",
            align="end",
            button_type="primary",
        )
        self.open_developer_link_button = pn.widgets.Button.from_param(
            self.param.open_developer_link,
            name="🔗 Open",
            width=125,
            sizing_mode="fixed",
            align="end",
            button_type="light",
        )
        self.download_converted_files = pn.widgets.FileDownload(
            callback=self._download_callback,
            filename="build.zip",
            width=125,
            button_type="light",
            sizing_mode="fixed",
            label="📁 Download",
            align="end",
        )

        return pn.Row(
            self.convert_button,
            # , self.open_developer_link_button, self.download_converted_files,
            # self.share_button, self.open_shared_link_button, self.copy_shared_link_button,
            sizing_mode="stretch_width",
            margin=(0, 5, 10, 5),
        )



import holoviews as hv
import hvplot.pandas
import panel as pn
from bokeh.sampledata.iris import flowers

pn.extension(sizing_mode="stretch_width")
hv.extension("bokeh")

accent_color = "#ff286e"


def get_plot():
    scatter = flowers.hvplot.scatter(
        x="sepal_length", y="sepal_width", c=accent_color, responsive=True, height=300
    )
    hist = flowers.hvplot.hist("petal_width", c=accent_color, responsive=True, height=300)

    scatter.opts(size=10)

    selection_linker = hv.selection.link_selections.instance()

    scatter = selection_linker(scatter)
    hist = selection_linker(hist)

    scatter.opts(tools=["hover"], active_tools=["box_select"])
    hist.opts(tools=["hover"], active_tools=["box_select"])

    return (scatter + hist).cols(1)


plot = get_plot()

pn.template.FastListTemplate(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Cross Filtering",
    header_background=accent_color,
    main=[plot],
).servable()



"""The Gallery component enables users to select a project from a list of projects"""
import string
from pathlib import Path
from typing import List

import panel as pn

from panel_sharing.models import Gallery as GalleryModel
from panel_sharing.models import Project
from panel_sharing.utils import set_directory


def _read_projects(path: Path):
    examples = []
    for folder in path.iterdir():
        if folder.is_dir():
            with set_directory(folder):
                project = Project.read()
            project.name = string.capwords(folder.name.replace("-", " "))
            examples.append(project)
    return sorted(examples, key=lambda x: x.name)


class Gallery(GalleryModel, pn.viewable.Viewer):
    """Enables users to select a project from a list of projects"""

    def __init__(self, examples: List[Project], **params):
        super().__init__()

        layout = pn.Column(
            pn.pane.Markdown(
                "## 🎁 Examples\nClick a button below to select an example or check out the "
                "<fast-anchor href='sharing_gallery' appearance='hypertext'>Gallery</fast-anchor>.",
                margin=0,
            ),
            sizing_mode="stretch_width",
        )

        self._examples_map = {example.name: example for example in examples}

        self.value = self._examples_map.get("Welcome", examples[0])

        for example in examples:
            button = pn.widgets.Button(name=example.name, button_type="light")

            button.on_click(self._click_handler)
            layout.append(button)

        self._panel = layout

    def _click_handler(self, event):
        self.value = self._examples_map[event.obj.name]

    def __panel__(self):
        return self._panel

    @classmethod
    def read(cls, path: Path) -> "Gallery":
        """Returns a Gallery of projects read from the specified path"""
        examples = _read_projects(path)
        return cls(examples=examples)

    def get(self, name) -> Project:
        """Returns the project with the specified name"""
        return self._examples_map[name]


if __name__.startswith("bokeh"):
    pn.extension(template="fast")
    gallery = Gallery.read(Path(__file__).parent.parent / "examples")
    pn.Column(gallery.param.value, gallery).servable(target="sidebar")



"""A module of shared utilities"""
import logging
import os
import time
import urllib.parse as urlparse
from contextlib import ContextDecorator, contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, Optional

import panel as pn

APP_KEY_PARAMETER = "app"
PROJECT_KEY_PARAMETER = "project"
DEFAULT_APP = ""
EXAMPLE_KEY_PARAMETER = "example"
DEFAULT_EXAMPLE = "Welcome"


@contextmanager
def set_directory(path: Path):
    """Sets the cwd within the context

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """

    origin = Path().absolute()
    try:
        path.mkdir(parents=True, exist_ok=True)
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


def exception_handler(ex):
    """A general exception handler for panel apps"""
    logging.exception("Error", exc_info=ex)
    if pn.state.notifications and ex:
        pn.state.notifications.error(f"Error. {ex}")


def _get_app_key(key: str, session_args: Dict, default=""):
    if not key in session_args:
        return default
    try:
        return session_args[key][0].decode(encoding="utf8")
    except:  # pylint: disable=bare-except
        return default


def get_app_key(default=DEFAULT_APP):
    """Returns the value of the app session arg or the default value"""
    return _get_app_key(key=APP_KEY_PARAMETER, session_args=pn.state.session_args, default=default)


def get_project_key(default=""):
    """Returns the value of the project session arg or the default value"""
    return _get_app_key(
        key=PROJECT_KEY_PARAMETER, session_args=pn.state.session_args, default=default
    )


def get_example_key(default=DEFAULT_EXAMPLE):
    """Returns the value of the example session arg or the default value"""
    return _get_app_key(
        key=EXAMPLE_KEY_PARAMETER, session_args=pn.state.session_args, default=default
    )


def notify_app_key_not_found(app_or_example, key: str):
    """Helper function to notify the given key was not found on load"""
    notification = f"The {app_or_example} '{key}' was not found"

    def on_load():
        pn.state.notifications.error(notification)

    pn.state.onload(on_load)


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


@dataclass
class Timer(ContextDecorator):
    """Time your code using a class, context manager, or decorator"""

    # Source: https://realpython.com/python-timer/

    timers: ClassVar[Dict[str, float]] = {}
    name: Optional[str] = None
    text: str = "{}: Elapsed time: {:0.4f} seconds"
    logger: Optional[Callable[[str], None]] = print
    _start_time: Optional[float] = field(default=None, init=False, repr=False)

    def __post_init__(self) -> None:
        """Initialization: add timer to dict of timers"""
        if not self.name:
            self.name = "Timer"
        if self.name:
            self.timers.setdefault(self.name, 0)

    def start(self) -> None:
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError("Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self) -> float:
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        # Calculate elapsed time
        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None

        # Report elapsed time
        if self.logger:
            self.logger(self.text.format(self.name, elapsed_time))
        if self.name:
            self.timers[self.name] += elapsed_time

        return elapsed_time

    def __enter__(self) -> "Timer":
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, *exc_info: Any) -> None:
        """Stop the context manager timer"""
        self.stop()


def del_query_params(*args, period=None):
    """Deletes the query arguments"""
    if not args:
        args = ("code", "state", "example", "project", "app")
    if pn.state.location:
        query = pn.state.location.query_params
        for parameter in args:
            if parameter in query:
                del query[parameter]

        def later():
            pn.state.location.search = "?" + urlparse.urlencode(query)

        if not period:
            later()
        else:
            pn.state.add_periodic_callback(later, period=period, count=1)



"""The ShareProject component enables users to share the project"""
import panel as pn
import param

from panel_sharing.components.js_actions import JSActions
from panel_sharing.models import AppState

LOGIN_TEXT = """\
## 🌎 Share

Please login to enable sharing.
"""
LICENSE_TEXT = """\
## 🌎 Share

By clicking the *Share* button, I make my code **open source, free and MIT licensed**.
"""

RAW_CSS = """
#sidebar a.bk-btn.bk-btn-success {
    color: var(--neutral-foreground-rest);
    padding-top: 5px !important;
}
"""


class ShareProject(pn.viewable.Viewer):
    """Enables users to share the project"""

    app_state: AppState = param.ClassSelector(class_=AppState)
    js_actions: JSActions = param.ClassSelector(class_=JSActions)

    share = param.Event()
    shared_url: str = param.String()

    open_shared_link: bool = param.Event()
    copy_shared_link: bool = param.Event()

    reset = param.Action()

    def __init__(self, app_state: AppState, js_actions: JSActions):
        super().__init__(app_state=app_state, js_actions=js_actions)
        self.reset = self._reset
        self.share_button = pn.widgets.Button.from_param(
            self.param.share,
            name="Share",
            sizing_mode="stretch_width",
            align="end",
            button_type="primary",
        )
        self.project = pn.widgets.TextInput.from_param(
            self.app_state.project.param.name, name="Project"
        )
        self.open_shared_link_button = pn.widgets.Button.from_param(
            self.param.open_shared_link,
            name="🔗 Open",
            sizing_mode="stretch_width",
            align="end",
            button_type="success",
        )
        self.copy_shared_link_button = pn.widgets.Button.from_param(
            self.param.copy_shared_link,
            name="✂️ Copy",
            sizing_mode="stretch_width",
            align="end",
            button_type="success",
        )
        self.download_shared_files_button = pn.widgets.FileDownload(
            callback=self._download_callback,
            filename="build.zip",
            button_type="success",
            sizing_mode="stretch_width",
            height=30,
            label="📁 Download",
            align="end",
        )
        if not RAW_CSS in pn.config.raw_css:
            pn.config.raw_css.append(RAW_CSS)

    def _download_callback(self):
        return self.app_state.project.to_zip_folder()

    @pn.depends("share", watch=True)
    def _share(self):
        self.shared_url = self.app_state.share()
        if pn.state.notifications:
            pn.state.notifications.success("Release succeeded")
        if pn.state.location:
            key = self.app_state.shared_key
            pn.state.location.search = ""
            pn.state.location.update_query(app=key)

    @pn.depends("open_shared_link", watch=True)
    def _open_shared_link(self):
        self.js_actions.open(url=self.shared_url)

    @pn.depends("copy_shared_link", watch=True)
    def _copy_shared_link(self):
        self.js_actions.copy(url=self.shared_url)

    @pn.depends("app_state.user.authenticated", "shared_url")
    def _panel(self):
        if not self.app_state.user.authenticated:
            return pn.Column(pn.pane.Markdown(LOGIN_TEXT, margin=0))
        if not self.shared_url:
            return pn.Column(
                pn.pane.Markdown(LICENSE_TEXT, margin=0),
                pn.widgets.TextInput.from_param(self.app_state.user.param.name, name="User"),
                self.project,
                self.share_button,
            )
        return pn.Column(
            pn.pane.Markdown(LICENSE_TEXT, margin=0),
            pn.widgets.TextInput.from_param(self.app_state.user.param.name, name="User"),
            self.project,
            self.share_button,
            self.copy_shared_link_button,
            self.open_shared_link_button,
            self.download_shared_files_button,
            sizing_mode="stretch_width",
        )

    def __panel__(self):
        return pn.panel(self._panel)

    def _reset(self, _):
        self.shared_url = ""


if __name__.startswith("bokeh"):
    pn.extension(notifications=True, template="fast")
    js_actions_ = JSActions()
    app = AppState()
    app.user.param.authenticated.constant = False
    share = ShareProject(app_state=app, js_actions=js_actions_)

    pn.Column(app.user.param.authenticated, share.param.reset, js_actions_).servable()
    share.servable(target="sidebar")



"""A set of components for Panel Sharing"""
from .build_project import ProjectBuilder
from .gallery import Gallery
from .oauth import OAuth
from .other import about, faq, iframe
from .project_gallery import create_project_gallery
from .settings_editor import settings_editor
from .share_project import ShareProject
from .source_editor import SourceEditor



import numpy as np
import panel as pn

pn.extension(sizing_mode="stretch_width")

INDICATORS = 32
PERIOD = 200  # mili seconds

indicators = (
    pn.indicators.Trend(
        data={"x": list(range(10)), "y": np.random.randn(10).cumsum()},
        width=150,
        height=100,
        plot_type=pn.indicators.Trend.param.plot_type.objects[i % 4],
    )
    for i in range(INDICATORS)
)
layout = pn.layout.FlexBox(*indicators)


def stream():
    for trend in layout:
        trend.stream(
            {"x": [trend.data["x"][-1] + 1], "y": [trend.data["y"][-1] + np.random.randn()]},
            rollover=20,
        )


periodic_callback = pn.state.add_periodic_callback(stream, PERIOD)

pn.template.FastListTemplate(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Streaming Indicators",
    main=[
        layout,
    ],
).servable()



"""Base models for Panel Sharing"""
# Should not contain any Panel UI elements
from __future__ import annotations

import base64
import json
import multiprocessing
import pathlib
import shutil
import tempfile
import uuid
from io import BytesIO
from pathlib import Path
from typing import Dict, List

import param
from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import BlobServiceClient, ContentSettings
from azure.storage.blob._list_blobs_helper import BlobPrefix
from panel import __version__
from panel.io.convert import convert_app

from panel_sharing import VERSION, config
from panel_sharing.utils import Timer, set_directory

ctx_forkserver = multiprocessing.get_context("forkserver")
ctx_forkserver.set_forkserver_preload(
    [
        # "base64",
        "bokeh",
        "holoviews",
        "hvplot",
        # "io",
        # "matplotlib",
        "numpy",
        "pandas",
        "panel",
        "param",
        # "PIL",
        # "skimage",
    ]
)

EXAMPLES = Path(__file__).parent / "examples"


def _convert_project(app: str = "source/app.py", dest_path: str = "build", requirements="auto"):
    """Helper function"""
    Path(dest_path).mkdir(parents=True, exist_ok=True)
    convert_app(app=Path(app), dest_path=Path(dest_path), requirements=requirements)  # type: ignore


class Source(param.Parameterized):
    """Represent the source files"""

    name = param.String(config.REPOSITORY_NAME, constant=True)
    code = param.String(config.CODE)
    readme = param.String(config.README)
    thumbnail = param.String(config.THUMBNAIL)
    requirements = param.String(config.REQUIREMENTS)

    def _items(self):
        """Returns the list of filename: text value"""
        return {
            "app.py": self.code,
            "readme.md": self.readme,
            "thumbnail.png": self.thumbnail,
            "requirements.txt": self.requirements,
        }.items()

    def save(self):
        """Saves the source files to the current working directory"""
        path = pathlib.Path()
        path.mkdir(parents=True, exist_ok=True)
        for file_path, text in self._items():
            pathlib.Path(path / file_path).write_text(text, encoding="utf8")

    @classmethod
    def read(cls) -> "Source":
        """Reads the Source from the current working directory"""
        path = pathlib.Path()
        source = cls(name="new")
        source.code = (path / "app.py").read_text()
        source.readme = (path / "readme.md").read_text()
        source.requirements = (path / "requirements.txt").read_text()
        return source

    def to_dict(self) -> Dict:
        """Returns the source as a dict"""
        return {
            "code": self.code,
            "readme": self.readme,
            # "thumbnail": self.thumbnail,
            "requirements": self.requirements,
        }


class Project(param.Parameterized):
    """A project consists of configuration and source files"""

    name = param.String(config.PROJECT_NAME)
    source = param.ClassSelector(class_=Source)
    files = param.Tuple(
        default=(
            "source/app.py",
            "source/requirements.txt",
            "source/readme.md",
            "source/thumbnail.png",
            "build/config.json",
            "build/app.html",
            "build/app.js",
        )
    )

    def __init__(self, **params):
        if "source" not in params:
            params["source"] = Source()
        super().__init__(**params)

        self._tmpdir = tempfile.TemporaryDirectory() # pylint: disable=consider-using-with
        self._tmppath = Path(self._tmpdir.name)
        self._save_hash = ""
        self._build_hash = ""

        if "name" not in params:
            with param.edit_constant(self):
                self.name = config.PROJECT_NAME

    def __del__(self):
        self._tmpdir.cleanup()

    def __str__(self):
        return self.name

    @classmethod
    def read(cls) -> "Project":
        """Reads the Project from the current working directory"""
        project = Project(name="new")
        with set_directory(Path("source")):
            project.source = Source.read()
        shutil.copytree(Path(), Path(project._tmpdir.name), dirs_exist_ok=True)
        project._save_hash=project._hash
        if Path("build").exists():
            project._build_hash=project._hash
        return project

    @property
    def _hash(self):
        return hash(json.dumps(self.to_dict()))

    def _reset_tmpdir(self):
        self._save_hash=""
        self._build_hash=""
        for child in self._tmppath.iterdir():
            if child.is_file():
                child.unlink()
            else:
                shutil.rmtree(child)

    def _save_to_tmpdir(self):
        _hash = self._hash
        if self._save_hash==_hash:
            return

        self._reset_tmpdir()
        with set_directory(self._tmppath/"source"):
            self.source.save()
        self._save_hash = _hash

    def save(self):
        """Saves the project files to the current working directory"""
        self._save_to_tmpdir()
        self._copy_from_tmpdir()

    def _get_requirements(self):
        requirements = pathlib.Path("source/requirements.txt")
        if requirements.exists() and requirements.read_text(encoding="utf8"):
            return str(requirements)
        return "auto"

    @property
    def _build_kwargs(self) -> Dict:
        """

        Assumes we are in the Project root and the source files are in ./source

        Returns:
            _description_
        """
        return {
            "app": "source/app.py",
            "dest_path": "build",
            "requirements": self._get_requirements(),
        }

    def save_build_json(self, kwargs: Dict):
        """Saves the build configuration in a json file in the current working directory"""
        if not kwargs:
            kwargs = self._build_kwargs

        build_json = {
            "app_builder": {"panel sharing": VERSION},
            "app_framework": {"panel": __version__},
            "build_kwargs": kwargs,
        }
        with open("config.json", "w", encoding="utf8") as file:
            json.dump(obj=build_json, fp=file, indent=1)

    def _remove_tmpbuilddir(self):
        self._build_hash=""
        tmpbuildpath = self._tmppath/"build"
        if tmpbuildpath.exists():
            shutil.rmtree(tmpbuildpath)

    def _build_to_tmpdir(self, base_target):
        _hash = self._hash + hash(base_target)
        if self._build_hash==_hash:
            return

        self._remove_tmpbuilddir()
        with set_directory(self._tmppath):
            kwargs = self._build_kwargs

            # We need to be really careful when we convert. See
            # https://github.com/holoviz/panel/issues/3939
            with Timer("convert project"):
                process = ctx_forkserver.Process(
                    target=_convert_project,
                    kwargs=kwargs,
                )
                process.start()

                with set_directory(Path("build")):
                    self.save_build_json(kwargs)
                process.join()
                if base_target != "":
                    app_html = Path("build/app.html")
                    text = app_html.read_text(encoding="utf8")
                    text = text.replace("<head>", "<head><base target='_blank' />")
                    app_html.write_text(text, encoding="utf8")

        self._build_hash = _hash

    def _copy_from_tmpdir(self):
        dst = str(Path().absolute())
        shutil.copytree(self._tmppath, dst, dirs_exist_ok=True)

    def build(self, base_target=""):
        """Saves and builds (i.e. converts) to the current working directory"""
        self._save_to_tmpdir()
        self._build_to_tmpdir(base_target=base_target)
        self._copy_from_tmpdir()

    def rebuild(self, base_target=""):
        """Forces as rebuild"""
        self._save_hash=""
        self._build_hash=""
        self.build(base_target=base_target)

    def to_dict(self):
        """Returns the project as a dictionary"""
        return {"source": self.source.to_dict()}

    def to_base64(self):
        """Returns the project base64 encoded"""
        value = json.dumps(self.to_dict(), separators=(",", ":"))
        result = base64.b64encode(value.encode(encoding="utf8")).decode(encoding="utf8")
        print("base64 len", len(result))
        return result

    @classmethod
    def from_dict(cls, value: Dict) -> "Project":
        """Returns a project from the value provided"""
        source = value.pop("source")
        return Project(source=Source(**source))

    @classmethod
    def from_base64(cls, value: str) -> "Project":
        """Returns a project from the base64 value"""
        value_dict = json.loads(base64.b64decode(value).decode(encoding="utf8"))
        return cls.from_dict(value_dict)

    def __eq__(self, other):
        return isinstance(other, Project) and self.to_dict() == other.to_dict()

    def to_zip_folder(self) -> BytesIO:
        """Returns the project as a .zip folder"""
        with tempfile.TemporaryDirectory() as tmpdir:
            with set_directory(pathlib.Path(tmpdir)):
                with set_directory(pathlib.Path("project")):
                    self.save()
                    self.build()
                target_file = "saved"
                result = shutil.make_archive(
                    target_file, "zip", root_dir=Path("project").absolute()
                )
                with open(result, "rb") as file:
                    return BytesIO(file.read())

    @staticmethod
    def from_zip_folder(zip_folder: BytesIO):
        """Creates a Project from a zip_folder"""
        with tempfile.TemporaryDirectory() as tmpdir:
            with set_directory(pathlib.Path(tmpdir)):
                target_file = "saved.zip"
                with open(target_file, "wb") as file:
                    file.write(zip_folder.getbuffer())
                    shutil.unpack_archive(target_file, format="zip")
                    return Project.read()

    def copy(self, project: Project):
        """Copies the given project including any saved and build files"""
        self.source.code = project.source.code
        self.source.readme = project.source.readme
        self.source.requirements = project.source.requirements
        self.source.thumbnail = project.source.thumbnail
        self._save_hash=""
        self._build_hash=""
        self._reset_tmpdir()

        # pylint: disable=protected-access
        with set_directory(self._tmppath):
            project._copy_from_tmpdir()
        self._save_hash=project._save_hash
        self._build_hash=project._build_hash




class User(param.Parameterized):
    """A User of the site"""

    name = param.String(config.GUEST_USER_NAME, constant=True, regex=config.USER_NAME_REGEX)
    authenticated = param.Boolean(config.AUTHENTICATED, constant=True)

    def __init__(self, **params):
        super().__init__(**params)

        if "name" not in params:
            with param.edit_constant(self):
                self.name = config.GUEST_USER_NAME

    def __str__(self):
        return self.name

    def authenticate(self, name):
        """Authenticates the give name"""
        with param.edit_constant(self):
            if name and name != config.GUEST_USER_NAME:
                self.name = name
                self.authenticated = True
            else:
                self.name = config.GUEST_USER_NAME
                self.authenticated = False


class Storage(param.Parameterized):
    """Represent a key-value where the value is a Project"""

    def __getitem__(self, key):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def keys(self):
        """Returns the list of keys of the storage"""
        raise NotImplementedError()

    def copy(self, key: str, project: Project):
        """Copy the source project to the specified key"""
        raise NotImplementedError()


class FileStorage(Storage):
    """The FileStorage represent a storage as files"""

    base_target = param.Selector(default="", objects=["", "_blank"])

    def __init__(self, path: str, **params):
        super().__init__(**params)

        self._path = pathlib.Path(path).absolute()

    def __getitem__(self, key):
        with set_directory(self._path / "projects" / key):
            project = Project.read()
        return project

    def _get_project_path(self, key) -> pathlib.Path:
        return self._path / "projects" / key

    def _get_www_path(self, key) -> pathlib.Path:
        return self._path / "www" / key

    def _move_locally(self, tmppath: pathlib.Path, project: pathlib.Path, www: pathlib.Path):
        if project.exists():
            shutil.rmtree(project)
        shutil.copytree(tmppath, project)
        if www.exists():
            shutil.rmtree(www)
        shutil.copytree(tmppath / "build", www)

    def __setitem__(self, key: str, value: Project):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = pathlib.Path(tmpdir)
            with set_directory(tmppath):
                value.save()
                value.build(base_target=self.base_target)

            project = self._get_project_path(key)
            www = self._get_www_path(key)

            self._move_locally(tmppath, project, www)

    def copy(self, key: str, project: Project):
        project_dir = self._get_project_path(key)
        www_dir = self._get_www_path(key)

        with tempfile.TemporaryDirectory() as source:
            source_path = Path(source)
            with set_directory(source_path):
                project.save()
                project.build()
                self._move_locally(source_path, project_dir, www_dir)

    def __delitem__(self, key):
        raise NotImplementedError()

    def keys(self):
        raise NotImplementedError()


class TmpFileStorage(FileStorage):
    """A FileStorage with temporary files that are cleaned up when no longer in use"""

    def __getitem__(self, key):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def keys(self):
        raise NotImplementedError()


class AzureBlobStorage(Storage):
    """An Azure Blob Storage"""

    base_target = param.Selector(default="", objects=["", "_blank"])
    blob_url = param.String(default=config.AZURE_BLOB_URL)
    web_url = param.String(default=config.AZURE_WEB_URL)
    project_container_name = param.String(default=config.AZURE_PROJECT_CONTAINER_NAME)
    web_container_name = param.String(default=config.AZURE_WEB_CONTAINER_NAME)
    conn_str = param.String(default=config.AZURE_BLOB_CONN_STR)

    def __init__(self, **params):
        super().__init__(**params)

        if not self.conn_str:
            raise ValueError("Error. No conn_str provided!")

        # Create the BlobServiceClient object
        self.service_client = BlobServiceClient.from_connection_string(self.conn_str)
        self.project_container_client = self.service_client.get_container_client(
            self.project_container_name
        )
        self.web_container_client = self.service_client.get_container_client(
            self.web_container_name
        )

    def __getitem__(self, key):
        with tempfile.TemporaryDirectory() as tmpdir:
            with set_directory(Path(tmpdir)):
                self._save_locally(key)
                return Project.read()

    def _save_locally(self, key: str):
        prefix = key + "/"

        project_generator = self.project_container_client.list_blobs(name_starts_with=prefix)
        for blob in project_generator:
            file = blob.name.replace(prefix, "")
            Path(file).parent.mkdir(parents=True, exist_ok=True)
            with open(file=file, mode="wb") as download_file:
                download_file.write(
                    self.project_container_client.download_blob(blob.name).readall()
                )

        web_generator = self.web_container_client.list_blobs(name_starts_with=prefix)
        for blob in web_generator:
            file = blob.name.replace(prefix, "build/")
            Path(file).parent.mkdir(parents=True, exist_ok=True)
            with open(file=file, mode="wb") as download_file:
                try:
                    download_file.write(
                        self.web_container_client.download_blob(blob.name).readall()
                    )
                except ResourceNotFoundError as ex:
                    raise Exception(
                        f"The container {self.web_container_name} or blob {blob} was not found"
                    ) from ex

    @staticmethod
    def _is_build_file(file: Path):
        return str(file).startswith("build/")

    def _get_container_name(self, file: Path):
        if self._is_build_file(file):
            return self.web_container_name
        return self.project_container_name

    def _get_file_path(self, file: Path):
        if self._is_build_file(file):
            return Path(file.name)
        return file

    def _get_blob(self, key: str, file: Path):
        file_path = self._get_file_path(file)
        return key + "/" + str(file_path)

    def _get_blob_client(self, key: str, file: Path):
        container_name = self._get_container_name(file)
        blob = self._get_blob(key, file)
        return self.service_client.get_blob_client(container=container_name, blob=blob)

    def _get_content_settings(self, file: Path) -> ContentSettings | None:
        if file.name.endswith("html"):
            return ContentSettings(content_type="text/html")
        return None

    def __setitem__(self, key, value: Project):
        with tempfile.TemporaryDirectory() as tmpdir:
            with set_directory(Path(tmpdir)):
                value.save()
                value.build(self.base_target)
                for file in Path().rglob("*"):
                    if file.is_file():
                        blob_client = self._get_blob_client(key=key, file=file)
                        content_settings = self._get_content_settings(file)
                        with open(file, mode="rb") as data:
                            blob_client.upload_blob(
                                data, overwrite=True, content_settings=content_settings
                            )

    def __delitem__(self, key):
        raise NotImplementedError()

    def keys(self):
        """Returns the list of keys of the storage"""
        raise NotImplementedError()

    def copy(self, key: str, project: Project):
        raise NotImplementedError()
        # project = self._get_project_path(key)
        # www = self._get_www_path(key)
        # self._move_locally(source, project, www)

    def delete(self, key):
        """Delete the key"""
        for file_ in Project.files:
            file = Path(file_)
            blob = self._get_blob(key=key, file=file)
            container_name = self._get_container_name(file)
            blob_client = self.service_client.get_blob_client(container=container_name, blob=blob)
            blob_client.delete_blob(snapshot=None)

    def get_url(self, key, file):
        """Returns the app url of the given key and file"""
        file = Path(file)
        container_name = self._get_container_name(file)
        file_path = self._get_file_path(file)
        if container_name == "$web":
            return self.web_url + key + "/" + str(file_path)
        return self.blob_url + container_name + "/" + key + "/" + str(file_path)


    def get_keys(self) -> List[str]:
        """Returns the app keys from the blob storage"""
        keys = []
        for user in self.web_container_client.walk_blobs('', delimiter='/'):
            if isinstance(user, BlobPrefix):
                for app in user:
                    if isinstance(app, BlobPrefix):
                        keys.append(app.name[:-1])
        return keys


class Site(param.Parameterized):
    """A site like awesome-panel.org. But could also be another site"""

    name: str = param.String(config.SITE, constant=True)
    title: str = param.String(config.TITLE, constant=True)
    faq: str = param.String(config.FAQ, constant=True)
    about: str = param.String(config.ABOUT, constant=True)
    thumbnail: str = param.String(config.THUMBNAIL, constant=True)

    development_storage: Storage = param.ClassSelector(class_=Storage, constant=True)
    examples_storage: Storage = param.ClassSelector(class_=Storage, constant=True)
    production_storage: Storage = param.ClassSelector(class_=Storage, constant=True)

    auth_provider = param.Parameter(constant=True)

    def __init__(self, **params):
        if "development_storage" not in params:
            params["development_storage"] = TmpFileStorage(path="apps/dev", base_target="_blank")
        if "examples_storage" not in params:
            params["examples_storage"] = FileStorage(path="apps/examples")
        if "production_storage" not in params:
            params["production_storage"] = FileStorage(path="apps/prod")

        super().__init__(**params)

        if "name" not in params:
            with param.edit_constant(self):
                self.name = config.SITE

    def get_shared_key(self, user: User, project: Project) -> str:
        """Returns the key of the user and project"""
        return user.name + "/" + project.name

    def get_shared_src(self, key):
        """Returns the shared url"""
        if isinstance(self.production_storage, AzureBlobStorage):
            return self.production_storage.get_url(key, "build/app.html")
        return f"apps/{key}/app.html"

    def get_development_src(self, key):
        """Returns the development url"""
        return f"apps-dev/{key}/app.html"


class AppState(param.Parameterized):
    """Represents the state of the Sharing App"""

    site: Site = param.Parameter(constant=True)
    user: User = param.Parameter(constant=True)
    project: Project = param.Parameter(constant=True)

    development_key: str = param.String()
    development_url: str = param.String()

    shared_key: str = param.String()
    shared_url: str = param.String()

    examples: pathlib.Path = param.ClassSelector(default=EXAMPLES, class_=pathlib.Path)

    def __init__(self, **params):
        if "site" not in params:
            params["site"] = Site()
        if "user" not in params:
            params["user"] = User()
        if "project" not in params:
            params["project"] = Project()

        super().__init__(**params)

    def _set_development(self, key: str):
        self.development_key = key
        if not key:
            self.development_url = ""
        else:
            self.development_url = self.site.get_development_src(key)

    def set_shared(self, key: str):
        """Sets the shared_key and shared_url"""
        self.shared_key = key
        self.shared_url = self.site.get_shared_src(key)

    @param.depends("user.name", "project.name", watch=True, on_init=True)
    def _update_shared(self):
        key = self.site.get_shared_key(user=self.user, project=self.project)
        self.set_shared(key)

    def _get_random_key(self):
        return str(uuid.uuid4())

    def copy(self, project: Project):
        """Copies the project from the source path"""
        self.project.copy(project)

        key = self._get_random_key()
        self.site.development_storage.copy(key=key, project=self.project)
        self._set_development(key)

    def build(self):
        """Build the current project and reload the app"""
        # We need to use a new key to trigger the iframe to refresh
        # The panel server somehow messes with the file
        self._set_development("")
        key = self._get_random_key()
        self.site.development_storage[key] = self.project
        self._set_development(key)

    def share(self):
        """Shared the current project"""
        key = self.shared_key
        url = self.shared_url
        self.site.production_storage[key] = self.project
        return url

    def login(self):
        """Logs the user in"""
        with param.edit_constant(self.user):
            self.user.authenticated = True

    def logout(self):
        """Logs the user out"""
        with param.edit_constant(self.user):
            self.user.authenticated = False

    def set_dev_project_from_shared_app(self, key):
        """Set the current project from an app key"""
        project = self.site.production_storage[key]

        self.project.source.code = project.source.code
        self.project.source.readme = project.source.readme
        self.project.source.requirements = project.source.requirements

        # pylint: disable=protected-access
        key = self._get_random_key()

        self.site.development_storage.copy(key=key, project=project)
        self._set_development(key)


class Gallery(param.Parameterized):
    """Represents a Gallery of Projects"""

    value = param.ClassSelector(class_=Project)



"""The settings_editor enables a user to edit some of the app settings"""
import panel as pn


def settings_editor(state):
    """The settings_editor enables a user to edit some of the app settings"""
    return pn.Column(
        pn.widgets.TextInput.from_param(state.user.param.name, name="User"),
        pn.widgets.TextInput.from_param(state.project.param.name, name="Project"),
        pn.widgets.TextInput.from_param(state.param.shared_url, name="Url", disabled=True),
        pn.widgets.Button(name="🗑️ DELETE", disabled=True),
        name="Configure",
        sizing_mode="stretch_both",
    )



import numpy as np
import panel as pn
from matplotlib.figure import Figure

pn.extension(sizing_mode="stretch_width", template="fast")


@pn.cache()
def plot_output(bins=10) -> Figure:
    fig0 = Figure(figsize=(12, 6))
    ax0 = fig0.subplots()

    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(437)
    ax0.hist(x, bins, density=True)

    return fig0


bins_input = pn.widgets.IntSlider(name="bins", value=20, start=20, end=200, step=20)
plot_output = pn.bind(plot_output, bins=bins_input)

pn.Column(
    bins_input,
    pn.panel(plot_output, height=400, sizing_mode="stretch_both", align="center"),
    sizing_mode="stretch_both",
).servable()

pn.state.template.param.update(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Bokeh Matplotlib App",
)



"""The Gallery component enables users to select a project from a list of projects"""
import random
from typing import List

import panel as pn

from panel_sharing import config
from panel_sharing.models import AzureBlobStorage
from panel_sharing.utils import Timer

RAW_CSS = """
.cards-grid {
  --cards-grid-min-size: 16rem;
  display: grid;
  grid-template-columns: repeat(auto-fill, minmax(var(--cards-grid-min-size), 1fr));
  grid-gap: 2rem;
  list-style: none;
}
/* Style the counter cards */
.card {
  box-shadow: 0 4px 8px 0 rgba(0, 0, 0, 0.2);
  padding: 16px;
  text-align: center;
  background-color: #f1f1f1;
}
/* Presentational styles */
.card {
  padding: 0px;
}

.cards-grid {
    margin: 2rem;
    padding: 0px;
}

.avatar {
    vertical-align: middle;
    float: right;
    width: 30px;
    height: 30px;
    margin-top: 5px;
    margin-bottom: 10px;
    margin-left: 5px;
    border-radius: 50%;
}
.card-actions {
    margin-left: 1em;
    margin-right: 1em;
}
.card-action svg {
    vertical-align: middle;
    float: left;
    height: 20px;
    color: white;
    margin-top: 10px;
    margin-right: 10px;
    fill: var(--neutral-foreground-rest);
}
.card-action.github-action svg {
    color: var(--neutral-foreground-rest);
}
.card-image {
    height: 200px;
    width: 100%;
}
.card-content {
    padding: 10px 10px 10px;
    color: var(--neutral-foreground-rest);
}

*|*:link {
    text-decoration: none;
}

.card-text {
    height: 100px;
}
.card-header {
    height: 2em;
}
"""

SHARING_URL = config.SITE_URL
WEB_URL = config.AZURE_WEB_URL
# pylint: disable=line-too-long
image_urls = [
    "https://media.wired.com/photos/5eaca02f381a11d296a5adab/master/w_1600,c_limit/photo_space_hubble_1_stsci-h-p2016a-m.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/thumb/5/54/Rayleigh-Taylor_instability.jpg/280px-Rayleigh-Taylor_instability.jpg",
    "https://media.wired.com/photos/5a5bbb1be251803428fea2a9/master/pass/hiv-virus-visual-science-company.jpg",
    "https://blog.siggraph.org/wp-content/uploads/2019/07/Art-of-Comm.png",
    "https://www.nvidia.com/content/dam/en-zz/Solutions/high-performance-computing/scientific-visualization/nvidia-vmd-arm-molecular-sim-4c25-p@2x.jpg",
]
# pylint: enable=line-too-long
MAX_CARDS = 50


def _get_card(key: str) -> str:
    """Returns a Fast HTML card"""
    app_author, app_name = key.split("/", 2)
    app_url = f"{WEB_URL}{key}/app.html"
    app_description = f"by {app_author}"
    app_author_url = f"https://github.com/{app_author}"
    app_code = f"sharing?app={key}"
    app_author_avatar = f"https://github.com/{app_author}.png?size=40"

    image_url = random.choice(image_urls)

    return f"""
<li class="card">
<fast-card class="gallery-item">
<a title="Click to open" class="card-action" href="{ app_url }">
<img class="card-image" src="{image_url}"/>
<div class="card-content">
    <h2 class="card-header">{ app_name }</h2>
    <p class="card-text">{ app_description }</p>
</div></a>
<div class="card-actions">
    <a class="card-action author-action" href="{ app_author_url }" target="_blank">
        <img src="{ app_author_avatar }" alt="avatar" class="avatar" title="author: {app_author}">
    </a>
    <a title="Edit the Code" appearance="neutral" class="card-action code-action" href="{app_code}">
        <svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg">
        <path fill-rule="evenodd" d="M4.854 4.146a.5.5 0 0 1 0 .708L1.707 8l3.147 3.146a.5.5 0 0 1-.708.708l-3.5-3.5a.5.5 0 0 1 0-.708l3.5-3.5a.5.5 0 0 1 .708 0zm6.292 0a.5.5 0 0 0 0 .708L14.293 8l-3.147 3.146a.5.5 0 0 0 .708.708l3.5-3.5a.5.5 0 0 0 0-.708l-3.5-3.5a.5.5 0 0 0-.708 0zm-.999-3.124a.5.5 0 0 1 .33.625l-4 13a.5.5 0 0 1-.955-.294l4-13a.5.5 0 0 1 .625-.33z"/>
        </svg>
    </a>
</div>
</fast-card>
</li>
"""


def _get_content(key: str, apps: List[str]):
    content = ""
    filter_apps = []
    for app in sorted(apps):
        if key in app:
            filter_apps.append(app)

    if len(filter_apps) > MAX_CARDS:
        filter_apps = filter_apps[0:MAX_CARDS]
    for app in sorted(filter_apps):
        content += _get_card(app)

    content = f"""
    <style>{RAW_CSS}</style>
    <div id="cards">
    <ul class="cards-grid">
    {content}
    </ul>
    </div>
    """

    if len(filter_apps) != len(apps):
        # pylint: disable=line-too-long
        content = (
            f"<p style='margin-left:35px'>Too many apps found. Showing the first {len(filter_apps)} of {len(apps)} apps. "
            "<b>Please make your search more specific</b>.</p>" + content
        )
        # pylint: enable=line-too-long
    return content


@pn.cache(ttl=60)
def _get_keys():
    with Timer(name="get_keys"):
        return AzureBlobStorage().get_keys()


def create_project_gallery():
    """Returns an AppGallery wrapped in a Template"""
    apps = _get_keys()

    container = pn.pane.HTML(sizing_mode="stretch_both")
    key_input = pn.widgets.TextInput(
        name="Search", margin=(0, 0, 0, 35), max_width=500, sizing_mode="stretch_width"
    )

    def get_content(key, apps):
        container.object = _get_content(key, apps)

    get_content("", apps)
    pn.bind(get_content, key=key_input.param.value_input, apps=apps, watch=True)

    return pn.template.FastListTemplate(
        site=config.SITE,
        site_url="https://awesome-panel.org/sharing_gallery",
        title="Panel Sharing Gallery",
        main=[key_input, container],
        main_layout="",
    )


if __name__.startswith("bokeh"):
    create_project_gallery().servable()



"""The VideoStreamInterface provides an easy way to apply transforms to a video stream"""
import base64
import io
import time

import numpy as np
import panel as pn
import param
import PIL
import skimage
from PIL import Image, ImageFilter
from skimage import data, filters
from skimage.color.adapt_rgb import adapt_rgb, each_channel
from skimage.draw import rectangle
from skimage.exposure import rescale_intensity
from skimage.feature import Cascade

pn.extension("terminal", sizing_mode="stretch_width")

HEIGHT = 400
WIDTH = 400
TIMEOUT = 250


def to_instance(value, **params):
    """Converts the value to an instance

    Args:
        value: A param.Parameterized class or instance

    Returns:
        An instance of the param.Parameterized class
    """
    if isinstance(value, param.Parameterized):
        value.param.update(**params)
        return value
    return value(**params)


class Timer(pn.viewable.Viewer):
    """Helper Component used to show duration trends"""

    _trends = param.Dict()

    def __init__(self, **params):
        super().__init__()

        self.last_updates = {}
        self._trends = {}

        self._layout = pn.Row(**params)

    def time_it(self, name, func, *args, **kwargs):
        """Measures the duration of the execution of the func function and reports it under the
        name specified"""
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        duration = round(end - start, 2)
        self._report(name=name, duration=duration)
        return result

    def inc_it(self, name):
        """Measures the duration since the last time `inc_it` was called and reports it under the
        specified name"""
        start = self.last_updates.get(name, time.time())
        end = time.time()
        duration = round(end - start, 2)
        self._report(name=name, duration=duration)
        self.last_updates[name] = end

    def _report(self, name, duration):
        if not name in self._trends:
            self._trends[name] = pn.indicators.Trend(
                title=name,
                data={"x": [1], "y": [duration]},
                height=100,
                width=150,
                sizing_mode="fixed",
            )
            self.param.trigger("_trends")
        else:
            trend = self._trends[name]
            next_x = max(trend.data["x"]) + 1
            trend.stream({"x": [next_x], "y": [duration]}, rollover=10)

    @pn.depends("_trends")
    def _panel(self):
        self._layout[:] = list(self._trends.values())
        return self._layout

    def __panel__(self):
        return self._panel


class ImageTransform(pn.viewable.Viewer):
    """Base class for image transforms."""

    def __init__(self, **params):
        super().__init__(**params)

        with param.edit_constant(self):
            self.name = self.__class__.name.replace("Transform", "")
        self.view = self.create_view()

    def __panel__(self):
        return self.view

    def run(self, image: str, height: int = HEIGHT, width: int = WIDTH) -> str:
        """Transforms the base64 encoded jpg image to a base64 encoded jpg BytesIO object"""
        raise NotImplementedError()

    def create_view(self):
        """Creates a view of the parameters of the transform to enable the user to configure them"""
        return pn.Param(self, name=self.name)

    def transform(self, image):
        """Transforms the image"""
        raise NotImplementedError()


class PILImageTransform(ImageTransform):
    """Base class for PIL image transforms"""

    @staticmethod
    def to_pil_img(value: str, height=HEIGHT, width=WIDTH):
        """Converts a base64 jpeg image string to a PIL.Image"""
        encoded_data = value.split(",")[1]
        base64_decoded = base64.b64decode(encoded_data)
        image = Image.open(io.BytesIO(base64_decoded))
        image.draft("RGB", (height, width))
        return image

    @staticmethod
    def from_pil_img(image: Image):
        """Converts a PIL.Image to a base64 encoded JPG BytesIO object"""
        buff = io.BytesIO()
        image.save(buff, format="JPEG")
        return buff

    def run(self, image: str, height: int = HEIGHT, width: int = WIDTH) -> io.BytesIO:
        pil_img = self.to_pil_img(image, height=height, width=width)

        transformed_image = self.transform(pil_img)

        return self.from_pil_img(transformed_image)

    def transform(self, image: PIL.Image) -> PIL.Image:
        """Transforms the PIL.Image image"""
        raise NotImplementedError()


class NumpyImageTransform(ImageTransform):
    """Base class for np.ndarray image transforms"""

    @staticmethod
    def to_np_ndarray(image: str, height=HEIGHT, width=WIDTH) -> np.ndarray:
        """Converts a base64 encoded jpeg string to a np.ndarray"""
        pil_img = PILImageTransform.to_pil_img(image, height=height, width=width)
        return np.array(pil_img)

    @staticmethod
    def from_np_ndarray(image: np.ndarray) -> io.BytesIO:
        """Converts np.ndarray jpeg image to a jpeg BytesIO instance"""
        if image.dtype == np.dtype("float64"):
            image = (image * 255).astype(np.uint8)
        pil_img = PIL.Image.fromarray(image)
        return PILImageTransform.from_pil_img(pil_img)

    def run(self, image: str, height: int = HEIGHT, width: int = WIDTH) -> io.BytesIO:
        np_array = self.to_np_ndarray(image, height=height, width=width)

        transformed_image = self.transform(np_array)

        return self.from_np_ndarray(transformed_image)

    def transform(self, image: np.ndarray) -> np.ndarray:
        """Transforms the nd.array image"""
        raise NotImplementedError()


class VideoStreamInterface(pn.viewable.Viewer):
    """An easy to use interface for a VideoStream and a set of transforms"""

    video_stream = param.ClassSelector(
        class_=pn.widgets.VideoStream, constant=True, doc="The source VideoStream"
    )

    height = param.Integer(
        HEIGHT,
        bounds=(10, 2000),
        step=10,
        doc="""The height of the image converted and shown""",
    )
    width = param.Integer(
        WIDTH,
        bounds=(10, 2000),
        step=10,
        doc="""The width of the image converted and shown""",
    )

    transform = param.Selector(doc="The currently selected transform")

    def __init__(
        self,
        transforms,
        timeout=TIMEOUT,
        paused=False,
        **params,
    ):
        super().__init__(
            video_stream=pn.widgets.VideoStream(
                name="Video Stream",
                timeout=timeout,
                paused=paused,
                height=0,
                width=0,
                visible=False,
                format="jpeg",
            ),
            **params,
        )
        self.image = pn.pane.JPG(height=self.height, width=self.width, sizing_mode="fixed")
        self._updating = False
        transforms = [to_instance(transform) for transform in transforms]
        self.param.transform.objects = transforms
        self.transform = transforms[0]
        self.timer = Timer(sizing_mode="stretch_width")
        self.settings = self._create_settings()
        self._panel = self._create_panel()

    def _create_settings(self):
        return pn.Column(
            pn.Param(
                self.video_stream,
                parameters=["timeout", "paused"],
                widgets={
                    "timeout": {
                        "widget_type": pn.widgets.IntSlider,
                        "start": 10,
                        "end": 2000,
                        "step": 10,
                    }
                },
            ),
            self.timer,
            pn.Param(self, parameters=["height", "width"], name="Image"),
            pn.Param(
                self,
                parameters=["transform"],
                expand_button=False,
                expand=False,
                widgets={
                    "transform": {
                        "widget_type": pn.widgets.RadioButtonGroup,
                        "orientation": "vertical",
                        "button_type": "success",
                    }
                },
                name="Transform",
            ),
            self._get_transform,
        )

    def _create_panel(self):
        return pn.Row(
            self.video_stream,
            pn.layout.HSpacer(),
            self.image,
            pn.layout.HSpacer(),
            sizing_mode="stretch_width",
            align="center",
        )

    @pn.depends("height", "width", watch=True)
    def _update_height_width(self):
        self.image.height = self.height
        self.image.width = self.width

    @pn.depends("transform")
    def _get_transform(self):
        # Hack: returning self.transform stops working after browsing the transforms for a while
        return self.transform.view

    def __panel__(self):
        return self._panel

    @pn.depends("video_stream.value", watch=True)
    def _handle_stream(self):
        if self._updating:
            return

        self._updating = True
        if self.transform and self.video_stream.value:
            value = self.video_stream.value
            try:
                image = self.timer.time_it(
                    name="Transform",
                    func=self.transform.run,
                    image=value,
                    height=self.height,
                    width=self.width,
                )
                self.image.object = image
            except PIL.UnidentifiedImageError:
                print("unidentified image")

            self.timer.inc_it("last update")
        self._updating = False


class GaussianBlur(PILImageTransform):
    """Gaussian Blur

    https://pillow.readthedocs.io/en/stable/reference/ImageFilter.html#PIL.ImageFilter.GaussianBlur
    """

    radius = param.Integer(default=2, bounds=(0, 10))

    def transform(self, image: Image):
        return image.filter(ImageFilter.GaussianBlur(radius=self.radius))


class GrayscaleTransform(NumpyImageTransform):
    """GrayScale transform

    https://scikit-image.org/docs/0.15.x/auto_examples/color_exposure/plot_rgb_to_gray.html
    """

    def transform(self, image: np.ndarray):
        grayscale = skimage.color.rgb2gray(image[:, :, :3])
        return skimage.color.gray2rgb(grayscale)


class SobelTransform(NumpyImageTransform):
    """Sobel Transform

    https://scikit-image.org/docs/0.15.x/auto_examples/color_exposure/plot_adapt_rgb.html
    """

    def transform(self, image):
        @adapt_rgb(each_channel)
        def sobel_each(image):
            return filters.sobel(image)

        return rescale_intensity(1 - sobel_each(image))


@pn.cache()
def get_detector():
    """Returns the Cascade detector"""
    trained_file = data.lbp_frontal_face_cascade_filename()
    return Cascade(trained_file)


class FaceDetectionTransform(NumpyImageTransform):
    """Face detection using a cascade classifier.

    https://scikit-image.org/docs/0.15.x/auto_examples/applications/plot_face_detection.html
    """

    scale_factor = param.Number(1.4, bounds=(1.0, 2.0), step=0.1)
    step_ratio = param.Integer(1, bounds=(1, 10))
    size_x = param.Range(default=(60, 322), bounds=(10, 500))
    size_y = param.Range(default=(60, 322), bounds=(10, 500))

    def transform(self, image):
        detector = get_detector()
        detected = detector.detect_multi_scale(
            img=image,
            scale_factor=self.scale_factor,
            step_ratio=self.step_ratio,
            min_size=(self.size_x[0], self.size_y[0]),
            max_size=(self.size_x[1], self.size_y[1]),
        )

        for patch in detected:
            rrr, ccc = rectangle(
                start=(patch["r"], patch["c"]),
                extent=(patch["height"], patch["width"]),
                shape=image.shape[:2],
            )
            image[rrr, ccc, 0] = 200

        return image


component = VideoStreamInterface(
    transforms=[
        GaussianBlur,
        GrayscaleTransform,
        SobelTransform,
        FaceDetectionTransform,
    ]
)

pn.template.FastListTemplate(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="VideoStream with transforms",
    sidebar=[component.settings],
    main=[component],
).servable()



"""The OAuth component enables users to authenticate with Github"""
from __future__ import annotations

import logging
import os
import uuid

import panel as pn
import param
import requests
from diskcache import Cache
from tornado.web import create_signed_value, decode_signed_value

from panel_sharing.utils import del_query_params

cache = Cache(".cache/panel_sharing_oauth")

logger = logging.getLogger("oauth")
logger.setLevel(logging.INFO)

SECURE_COOKIE = "state"
SECURE_COOKIE_TTL_IN_DAYS = 10


class JSActions(pn.reactive.ReactiveHTML):  # pylint: disable=too-many-ancestors
    """Helper class for triggering js functions"""

    # .event cannot trigger on js side. Thus I use Integer
    _url = param.String()
    _open = param.Boolean()

    _set_cookie = param.Dict()
    _delete_secure_cookie = param.String()

    _template = """<div id="jsaction" style="height:0px;width:0px"></div>"""
    _scripts = {
        "_open": "window.open(data._url, '_self')",
        "_set_cookie": """
function createCookie(name,value,days) {
    if (days) {
        var date = new Date();
        date.setTime(date.getTime()+(days*24*60*60*1000));
        var expires = "; expires="+date.toGMTString();
    }
    else var expires = "";
    document.cookie = name+"="+value+expires+"; path=/;secure";
}      
const {name, value, days}=data._set_cookie
createCookie(name, value, days)""",
        "_delete_secure_cookie": """
value=data._delete_secure_cookie+'=; Max-Age=-99999999; path=/;secure'
document.cookie = value
""",
    }

    def __init__(self):
        super().__init__(height=0, width=0, sizing_mode="fixed", margin=0)

    def open(self, url: str):
        """Opens the url in the same tab"""
        logger.info("Open and redirect to %s", url)
        if url:
            self._url = url
            self._open = not self._open
        else:
            raise ValueError("No url to open")

    def set_secure_cookie(self, name, value, days=1.0):
        """Sets a cookie as a secure cookie

        Args:
            name: The name of the cookie
            value: The value of the cookie. Please not you will have to encrypt this your self
            days: Days to expiration. Defaults to 1.0.
        """
        logger.info("set_secure_cookie %s %s", name, value)
        self._set_cookie = {"name": name, "value": value, "days": days}

    def delete_secure_cookie(self, name):
        """Deletes the cookie

        Args:
            name: The name of the cookie to delete
        """
        logger.info("delete_secure_cookie %s", name)
        self._delete_secure_cookie = name


class OAuth(pn.viewable.Viewer):
    """Enables users to authenticate with Github

    To use it please create an oauth app as described in
    https://docs.github.com/en/developers/apps/building-oauth-apps/creating-an-oauth-app

    Then follow the description in https://panel.holoviz.org/user_guide/Authentication.html and

    - Set the ENVIRONMENT VARIABLES
      - `PANEL_OAUTH_KEY`
      - `PANEL_OAUTH_SECRET`
      - `PANEL_OAUTH_REDIRECT_URI` (optional)
    - Create a Cookie Secret via `panel secret` and set the environment variable
      - `PANEL_COOKIE_SECRET`
    """

    log_in = param.Event(label="Log in via Github")
    log_out = param.Event(
        label="Log out",
    )

    user_info = param.Dict(constant=True)
    user = param.String(constant=True)

    state = param.String()

    def __init__(self):
        super().__init__()

        self._client_id = os.environ.get("PANEL_OAUTH_KEY", "")
        self._client_secret = os.environ.get("PANEL_OAUTH_SECRET", "")
        self._cookie_secret = os.environ.get("PANEL_COOKIE_SECRET", "")
        self._oauth_redirect_uri = os.environ.get("PANEL_OAUTH_REDIRECT_URI", "")
        self._jsactions = JSActions()

        self._set_user_from_state()
        if not self.user:
            self._set_user_from_redirect()

        pn.state.onload(lambda: del_query_params("code", "state", period=1000))

    @pn.depends("user")
    def _panel(self):
        if not self.configured:
            return pn.pane.Alert("Auth is not configured", alert_type="danger")
        if self.user:
            return pn.Column(
                self._jsactions,
                self.param.log_out,
                # pn.widgets.TextInput.from_param(self.param.user),
            )
        return pn.Column(
            self._jsactions, pn.widgets.Button.from_param(self.param.log_in, button_type="default")
        )

    @property
    def configured(self) -> bool:
        """Returns True if the component has been configured, i.e. the relevant environment
        variables have been set"""
        return all(
            (self._client_id, self._client_secret, self._cookie_secret, self._oauth_redirect_uri)
        )

    def __panel__(self):
        return pn.panel(self._panel)

    def _set_user_from_state(self):
        state = self._get_secure_cookie(SECURE_COOKIE)
        user_info = cache.get(state, {}).get("user_info", {})
        user = user_info.get("login", "")
        if user_info and user:
            with param.edit_constant(self):
                self.user_info = user_info
                self.user = user
                logger.info("Logged in user %s from user_info in state", self.user)
        else:
            logger.info("No user_info in state.")

    def _set_secure_cookie(self, name, value, days=10):
        signed_value = create_signed_value(self._cookie_secret, name, value).decode("utf8")
        self._jsactions.set_secure_cookie(name=name, value=signed_value, days=days)

    def _get_secure_cookie(self, name: str) -> str | None:
        signed_value = pn.state.cookies.get(name, "")
        value = decode_signed_value(secret=self._cookie_secret, name=name, value=signed_value)
        if value:
            return value.decode(encoding="utf8")
        return ""

    def _delete_secure_cookie(self, name: str):
        self._jsactions.delete_secure_cookie(name)

    def _is_valid(self, state: str) -> bool:
        cookie_state = self._get_secure_cookie("state")
        logger.info("expected state %s", state)
        logger.info("cookie   state %s", cookie_state)
        return bool(state and cookie_state == state)

    def _create_state(self):
        with param.edit_constant(self):
            self.state = str(uuid.uuid4())
        self._set_secure_cookie(SECURE_COOKIE, self.state, days=SECURE_COOKIE_TTL_IN_DAYS)
        cache[self.state] = {}
        return self.state

    @pn.depends("log_in", watch=True)
    def _log_in_handler(self, _=None):
        logger.info("Login triggered. Redirecting to https://github.com/login/oauth/authorize")
        url = (
            "https://github.com/login/oauth/authorize"
            f"?client_id={self._client_id}"
            f"&state={self._create_state()}"
            f"&allow_signup=true"
        )
        self._jsactions.open(url)

    @pn.depends("log_out", watch=True)
    def _log_out_handler(self):
        user = self.user
        state = self._get_secure_cookie(SECURE_COOKIE)
        self._delete_secure_cookie(SECURE_COOKIE)
        del cache[state]

        with param.edit_constant(self):
            self.user_info = {}
            self.user = ""
            self.param.trigger("user")
        logger.info("Logged out %s", user)

    def _set_user_from_redirect(self):
        code = self._get_code_from_session_args()
        if not code:
            logger.info("No code found in session_args")
            return
        logger.info("Code %s found in session_args", code)

        state = self.get_state_from_session_args()
        if not state:
            logger.info("No state found in session_args")
            return
        logger.info("State %s found in session_args", state)

        if self._is_valid(state=state):
            logger.info("State is valid")
            access_token = self._request_access_token(code)
            if access_token:
                logger.info("access_token received")
                user_info = self._request_user_info(access_token=access_token)
                if user_info:
                    logger.info("user_info received")
                    self._set_user(user_info, state=state)
                else:
                    logger.info("No user_info received")
            else:
                logger.info("No access_token received")
        else:
            logger.info("state is not valid")

    def _set_user(self, user_info, state):
        with param.edit_constant(self):
            self.user_info = user_info
            self.user = self.user_info["login"]
            cache[state] = {"user_info": self.user_info}

        logger.info("Logged in %s successfully", self.user)

    def _get_code_from_session_args(self):
        return pn.state.session_args.get("code", [b""])[0].decode("utf8")

    def get_state_from_session_args(self):
        """Returns the state value or '' if not given"""
        return pn.state.session_args.get("state", [b""])[0].decode("utf8")

    def _request_user_info(self, access_token: str):
        response_json = requests.get(
            url="https://api.github.com/user",
            timeout=20,
            headers=dict(
                authorization=f"Bearer {access_token}",
                accept="application/json",
            ),
        ).json()
        if "login" in response_json:
            return response_json
        return {}

    def _request_access_token(self, code: str):
        if not code:
            return ""

        access_token_json = requests.post(
            url="https://github.com/login/oauth/access_token",
            timeout=20,
            headers=dict(
                accept="application/json",
            ),
            data=dict(
                redirect_uri=self._oauth_redirect_uri,
                client_id=self._client_id,
                client_secret=self._client_secret,
                code=code,
            ),
        ).json()
        if not "access_token" in access_token_json:
            return ""
        return access_token_json["access_token"]


if __name__.startswith("bokeh"):
    print("RESTARTING")

    pn.extension(sizing_mode="stretch_width")

    component = OAuth()

    pn.Column(
        component,
        pn.bind(pn.pane.JSON, object=component.param.user_info),
    ).servable()



"""The SourceEditor enables users to edit an instance of the Source model"""
import panel as pn
import param

from panel_sharing.models import Source


class SourceEditor(pn.viewable.Viewer):
    """An Editor to edit an instance of the Source model"""

    source = param.Parameter(constant=True)

    def __init__(self, source: Source):
        super().__init__(source=source)
        self._panel = self._get_panel()

    def __panel__(self):
        return self._panel

    def _get_panel(self):
        code_tab = pn.widgets.Ace.from_param(
            self.source.param.code,
            language="python",
            theme="monokai",
            sizing_mode="stretch_both",
            name="app.py",
        )
        readme_tab = pn.widgets.Ace.from_param(
            self.source.param.readme,
            language="markdown",
            theme="monokai",
            sizing_mode="stretch_both",
            name="readme.md",
        )

        @pn.depends(dataurl=self.source.param.thumbnail)
        def thumbnail_tab(dataurl):
            return pn.pane.HTML(
                f"""<img src={dataurl} style="height:100%;width:100%"></img>""",
                max_width=700,
                name="thumbnail.png",
                sizing_mode="scale_width",
            )

        requirements_tab = pn.widgets.Ace.from_param(
            self.source.param.requirements,
            language="txt",
            theme="monokai",
            sizing_mode="stretch_both",
            name="requirements.txt",
        )
        return pn.Tabs(
            code_tab,
            readme_tab,
            ("thumbnail.png", thumbnail_tab),
            requirements_tab,
            sizing_mode="stretch_both",
        )



"""Create the Awesome Panel Sharing App"""
from __future__ import annotations

import logging
from pathlib import Path

import panel as pn
from diskcache import Cache

from panel_sharing import components
from panel_sharing.models import AppState, AzureBlobStorage, Project, Site
from panel_sharing.utils import (
    exception_handler,
    get_app_key,
    get_example_key,
    get_project_key,
    notify_app_key_not_found,
)

logger = logging.getLogger("panel_sharing")
logger.setLevel(logging.INFO)

RAW_CSS = """
.sidenav a {
    padding: 0px !important;
}
.sidenav a:hover {
    color: var(--accent-foreground-hover) !important;   
}
"""


cache = Cache(".cache/panel_sharing")


@pn.cache
def get_examples(examples: str):
    """Returns a list of example Projects"""
    # pylint: disable=protected-access
    return list(components.Gallery.read(Path(examples))._examples_map.values())


def _set_start_project(state, set_example, gallery):
    key = get_app_key()
    project = get_project_key()
    example = get_example_key()
    if key:
        try:
            state.set_dev_project_from_shared_app(key)
        except:  # pylint: disable=bare-except
            notify_app_key_not_found("app", key)
            set_example(project=gallery.value)
            pn.state.location.search = ""
    elif project:
        try:
            project_ = Project.from_base64(project)
            state.project.source.code = project_.source.code
            state.project.source.requirements = project_.source.requirements
            state.project.source.readme = project_.source.readme
            state.project.source.thumbnail = project_.source.thumbnail
            pn.state.onload(state.build)
        except:  # pylint: disable=bare-except
            notify_app_key_not_found("project", key)
            set_example(project=gallery.value)
            pn.state.location.search = ""
    else:
        try:
            set_example(project=gallery.get(example))
        except:  # pylint: disable=bare-except
            notify_app_key_not_found("example", key)
            set_example(project=gallery.value)
            pn.state.location.search = ""


def create():
    """Returns an instance of the Panel Sharing app

    Args:
        examples: A path to a gallery of example projects. Defaults to EXAMPLES.

    """
    logger.info("CREATE STARTED")
    pn.config.raw_css.append(RAW_CSS)
    pn.extension(
        "ace", sizing_mode="stretch_width", notifications=True, exception_handler=exception_handler
    )

    site = Site(production_storage=AzureBlobStorage())
    state = AppState(site=site)

    gallery = components.Gallery(examples=get_examples(str(state.examples.absolute())))

    @pn.depends(gallery.param.value, watch=True)
    def set_example(project):
        project.build()
        state.copy(project)
        if pn.state.location:
            logger.info("set_example: updating location")
            pn.state.location.search = ""
            pn.state.location.update_query(example=project.name)

    _set_start_project(state, set_example, gallery)

    project_builder = components.ProjectBuilder(state=state)
    source_editor = components.SourceEditor(source=state.project.source)
    editor_tab = pn.Column(project_builder, source_editor, sizing_mode="stretch_both", name="Edit")

    source_pane = editor_tab

    target_pane = components.iframe(src=state.param.development_url)
    share_project = components.ShareProject(app_state=state, js_actions=project_builder.jsactions)

    authentication = components.OAuth()

    pn.bind(state.user.authenticate, name=authentication.param.user, watch=True)
    logger.info("authenticating user")
    state.user.authenticate(authentication.user)

    @pn.depends(authentication.param.state, watch=True)
    def handle_auth_state(state, app_state=state):
        logger.info("handle_auth_state")
        if not state:
            return
        cache[state] = {
            "source": app_state.project.source.to_dict(),
            "key": app_state.development_key,
        }

    oauth_state = authentication.get_state_from_session_args()
    if oauth_state and oauth_state in cache:
        state.project.source.param.update(**cache[oauth_state]["source"])
        state._set_development(cache[oauth_state]["key"])  # pylint: disable=protected-access

    template = pn.template.FastGridTemplate(
        site=state.site.name,
        title=state.site.title,
        theme_toggle=False,
        prevent_collision=True,
        save_layout=True,
        site_url="https://awesome-panel.org",
        favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",  # pylint: disable=line-too-long
        sidebar=[
            share_project,
            authentication,
            gallery,
            project_builder.jsactions,
        ],
    )
    template.main[0:5, 0:6] = source_pane
    template.main[0:5, 6:12] = target_pane
    logger.info("CREATE FINISHED")
    return template


if __name__ == "__main__":
    from panel_sharing.utils import Timer

    with Timer(name="First"):
        create()
    with Timer(name="Second"):
        create()
    with Timer(name="Third"):
        create()



import numpy as np
from bokeh.layouts import row
from bokeh.models import Column, ColumnDataSource
from bokeh.models.widgets import Slider
from bokeh.plotting import figure

# Set up data
N = 200
x = np.linspace(0, 4 * np.pi, N)
y = np.sin(x)
source = ColumnDataSource(data=dict(x=x, y=y))


# Set up plot
plot = figure(
    plot_height=400,
    title="my sine wave",
    tools="crosshair,pan,reset,save,wheel_zoom",
    x_range=[0, 4 * np.pi],
    y_range=[-2.5, 2.5],
)

plot.line("x", "y", source=source, line_width=3, line_alpha=0.6)
plot.title.text = "My sine wave"


# Set up widgets
offset = Slider(title="offset", value=0.0, start=-5.0, end=5.0, step=0.1)
amplitude = Slider(title="amplitude", value=1.0, start=-5.0, end=5.0)
phase = Slider(title="phase", value=0.0, start=0.0, end=2 * np.pi)
freq = Slider(title="frequency", value=1.0, start=0.1, end=5.1)


# Set up callbacks
def update_data(attrname, old, new):

    # Get the current slider values
    a = amplitude.value
    b = offset.value
    w = phase.value
    k = freq.value

    # Generate the new curve
    x = np.linspace(0, 4 * np.pi, N)
    y = a * np.sin(k * x + w) + b

    source.data = dict(x=x, y=y)


for w in [offset, amplitude, phase, freq]:
    w.on_change("value", update_data)


# Set up layouts and add to document
inputs = Column(offset, amplitude, phase, freq, sizing_mode="fixed")

bokeh_component = row(inputs, plot, sizing_mode="stretch_width")

# Set up Panel

import panel as pn

pn.extension(template="fast")
pn.panel(bokeh_component).servable()

pn.state.template.param.update(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Bokeh App",
)



import panel as pn

pn.extension(sizing_mode="stretch_width", template="fast")

pn.Column(
    pn.panel(
        """# Welcome to Panel Sharing ❤️

[Panel](https://panel.holoviz.org) is an open-source, data app framework that supports your workflow from exploration to
production. Panel is very popular in *real* science, engineering and finance. It can be used successfully in
any domain.

[Panel sharing](https://github.com/awesome-panel/panel-sharing) is **the fastest way to develop, [convert](https://panel.holoviz.org/user_guide/Running_in_Webassembly.html) and share Panel data apps**.

Panel Sharing was made with Panel! Check out the code and **report issues
[here](https://github.com/awesome-panel/panel-sharing)**.

**Select an example app in the sidebar** to get started.
"""
    ),
    pn.pane.Alert(
        "Panel Sharing is **currently a prototype**. **Your code and apps will not be persisted!**",
        alert_type="danger",
        margin=(0, 10),
    ),
).servable()

pn.state.template.param.update(
    site="Awesome Panel Sharing",
    site_url="https://awesome-panel.org/sharing",
    favicon="https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/320297ccb92773da099f6b97d267cc0433b67c23/favicon/ap-1f77b4.ico",
    title="Welcome",
)
